module Data.DataLoader exposing (Entry, decode)

import Csv
import Csv.Decode
import Utils exposing (lookupCountry)

{- Type to represent entries of the Schengen Visa Dataset.

Description:
- state: Schengen State
- consulate: Consulate of a Schengen State located in a third country
- consulateCountry: Third country in which consulate is located
- year: Year the entry was supplied for
- totalVisasApplied: Sum of Uniform visas that were applied for
- totalVisasIssued: Sum of Uniform visas that were issued
- totalAppliedMEV: Amount of Uniform Visas that were issued as MEVs

Additional Notes:
- Visas come in multiple forms, e.g. ATVs, LTVs and Uniform Visas
- ATV = Airport Transit Visa; needed for a stopover in the Schengen Region
when travelling from and to a Non-Schengen Region
- LTV = Limited Territorial Validity Visa; allows the travel to only
a subset of the States in the Schengen Region
- Uniform (Short Stay) Visa; allows for entry to and travel within the whole
Schengen Region for a specific time period
=> ATVs, LTVs and Uniform Visas can also be issued in the form of a MEV (Multiple Entry Visa),
which allows the holder to enter and exit the Schengen Region multiple times within the time period

-}
type alias Entry = 
    { state: String
    , consulate: String
    , consulateCountry: String
    , year: Int
    -- Note: These columns contain values like 5.0 which are parsed as float.
    -- Additional data sanity will be necessary.
    , totalVisasApplied: Maybe Float
    , totalVisasIssued: Maybe Float
    , totalIssuedMEV: Maybe Float
    }

override : List Entry -> List Entry
override entries =
    let
        stateToISO : Entry -> Entry
        stateToISO entry =
            { entry | state = lookupCountry entry.state |> .code }

        consulateToISO : Entry -> Entry
        consulateToISO entry =
            { entry | consulateCountry = lookupCountry entry.consulateCountry |> .code }
    in
    entries |> List.map stateToISO |> List.map consulateToISO

decode : String -> Result Csv.Decode.Errors (List Entry)
decode csv =
    Csv.parse csv
        |> Csv.Decode.decodeCsv decoder
        |> Result.map override

decoder : Csv.Decode.Decoder (Entry -> a) a
decoder =
    Csv.Decode.field "Schengen State" Ok
    |> Csv.Decode.andMap (Csv.Decode.field "Consulate" Ok)
    |> Csv.Decode.andMap (Csv.Decode.field "Country where consulate is located" Ok)
    |> Csv.Decode.andMap (String.toInt >> Result.fromMaybe "Error parsing Year" |> Csv.Decode.field "Year")
    |> Csv.Decode.andMap (String.toFloat >> Result.fromMaybe "Error parsing total of uniform visas applied" |> Csv.Decode.maybe |> Csv.Decode.field "Uniform visas applied for")
    |> Csv.Decode.andMap (String.toFloat >> Result.fromMaybe "Error parsing total of uniform visas issued" |> Csv.Decode.maybe |> Csv.Decode.field "Total  uniform visas issued (including MEV)")
    |> Csv.Decode.andMap (String.toFloat >> Result.fromMaybe "Error parsing total of multiple entry uniform visas issued" |> Csv.Decode.maybe |> Csv.Decode.field "Multiple entry uniform visas (MEVs) issued")
    |> Csv.Decode.map Entry