module MapLoader exposing (Polygon, Country, decoder, synonym, toCountries)

import Dict exposing (Dict)
import Json.Decode
import Json.Encode
import GeoJson exposing (FeatureObject, GeoJson, GeoJsonObject(..), Geometry(..), Position)
import Triple
import Utils exposing (isCodeValid, lookupCountry)

decoder : Json.Decode.Decoder GeoJson.GeoJson
decoder = GeoJson.decoder

type alias GeoProperty =
    { name: String
    , name_long: String
    , iso_a2: String
    , labelX: Float
    , labelY: Float
    }

synonym : Country -> String -> Bool
synonym country iso =
    country.meta.iso_a2 == iso

propertyDecoder : Json.Decode.Decoder GeoProperty
propertyDecoder =
    Json.Decode.map5 GeoProperty
        (Json.Decode.field "name" Json.Decode.string)
        (Json.Decode.field "name_long" Json.Decode.string)
        (Json.Decode.field "iso_a2" Json.Decode.string)
        (Json.Decode.field "label_x" Json.Decode.float)
        (Json.Decode.field "label_y" Json.Decode.float)

type alias Polygon = List ( Float, Float )
type alias Country =
    { meta: GeoProperty, polygons: List Polygon }

-- Gets rid of the variants in GeoJsonObject.
toFeatures : GeoJson -> List FeatureObject
toFeatures ( geoObj, bbox ) =
    case geoObj of
        Geometry geom ->
            [ { id = Nothing, properties = Json.Encode.null, geometry = Just geom } ]

        Feature feature ->
            [ feature ]

        FeatureCollection features ->
            features

toCountry : FeatureObject -> Maybe Country
toCountry obj =
    case Json.Decode.decodeValue propertyDecoder obj.properties of
        Err _ -> Nothing 
        Ok meta ->
            let
                polygons : List Polygon
                polygons = obj.geometry
                    |> Maybe.map toPolygons
                    |> Maybe.withDefault []
            in
            Country meta polygons |> Just


toCountries : GeoJson -> List Country
toCountries obj =
    obj |> toFeatures |> List.map toCountry |> List.filterMap identity |> override


-- Allows for manual overrides of country names s.t.
-- it might be consistent w.r.t. other data to merge with.
override : List Country -> List Country
override countries =
    let
        replace : Country -> Country
        replace country =
            if isCodeValid country.meta.iso_a2 then country
            else let meta = country.meta in
                { country | meta = { meta | iso_a2 = lookupCountry country.meta.name |> .code } }
    in 
    List.map replace countries

-- Transforms Geometry into a list of polygons.
-- Note: The Polygon variant in GeoJSON could actually comprise of more
-- than one polygon (which is confusing since MultiPolygon exists).
-- This happens because the format supports the description of "holes" (which we ignore here!) as polygons
-- within the first given polygon. See also https://stevage.github.io/geojson-spec/#appendix-A.3.
toPolygons : Geometry -> List Polygon
toPolygons geom =
    case geom of 
        GeoJson.Polygon polygons ->
            polygons |> List.map transform
        GeoJson.MultiPolygon multiPolygon ->
            multiPolygon |> List.concat |> List.map transform
        _ -> []

transform : List Position -> Polygon
transform positions =
    positions |> List.map Triple.dropThird
